import pytest
from tests.fixtures.collector import basic_collector, disk_collector

from ml_instrumentation.Collector import Collector

@pytest.mark.parametrize('collector_fixture', [basic_collector, disk_collector])
def test_benchmark_write_path1(collector_fixture, request, benchmark):
    collector: Collector = request.getfixturevalue(collector_fixture.__name__)
    collector.set_experiment_id(0)

    def _inner():
        for i in range(1_000):
            collector.next_frame()
            collector.collect('m1', i)
            collector.collect('m2', f'test string {i}')

        # force any in-flight data to be sync'd
        # normally would close() the collector, but collector
        # will be reused for this test
        collector._writer.sync_now()

    benchmark(_inner)

@pytest.mark.parametrize('collector_fixture', [basic_collector, disk_collector])
def test_benchmark_read1(collector_fixture, request, benchmark):
    collector: Collector = request.getfixturevalue(collector_fixture.__name__)
    collector.set_experiment_id(0)
    for i in range(1_000):
        collector.next_frame()
        collector.collect('m1', i)
        collector.collect('m2', f'test string {i}')

    collector._writer.sync_now()

    def _inner():
        d = collector.get('m1', 0)
        assert len(d) == 1_000

        d = collector.get('m2', 0)
        assert len(d) == 1_000

    benchmark(_inner)



from collections.abc import Callable
import numpy as np

class Sampler:
    def next(self, v: float) -> float | None: ...
    def next_eval(self, c: Callable[[], float]) -> float | None: ...
    def end(self) -> float | None: ...

class Ignore:
    def __init__(self): ...
    def next(self, v): return None
    def next_eval(self, v): return None
    def end(self): return None

# by definition, this must be a stateless object
# safe to instantiate a singleton instance for default parameters
ignore = Ignore()

class Identity(Sampler):
    def next(self, v: float):
        return v

    def next_eval(self, c: Callable[[], float]):
        return c()

    def end(self):
        return None

# by definition, this must be a stateless object
# safe to instantiate a singleton instance for default parameters
identity = Identity()

# --------------
# -- Samplers --
# --------------
class Window(Sampler):
    def __init__(self, size: int):
        self._b = np.empty(size, dtype=np.float64)
        self._clock = 0
        self._size = size

    def next(self, v: float):
        self._b[self._clock] = v
        self._clock += 1

        if self._clock == self._size:
            m = self._b.mean()
            self._clock = 0
            return m

    def next_eval(self, c: Callable[[], float]):
        return self.next(c())

    def end(self):
        out = None
        if self._clock > 0:
            out = self._b[:self._clock].mean()

        self._clock = 0
        return out

class Subsample(Sampler):
    def __init__(self, freq: int):
        self._clock = 0
        self._freq = freq

    def next(self, v: float):
        tick = self._clock % self._freq == 0
        self._clock += 1

        if tick:
            return v

    def next_eval(self, c: Callable[[], float]):
        tick = self._clock % self._freq == 0
        self._clock += 1

        if tick:
            return c()

    def end(self):
        self._clock = 0
        return None

class MovingAverage(Sampler):
    def __init__(self, decay: float):
        self._decay = decay
        self.z = 0.

    def next(self, v: float):
        self.z = self._decay * self.z + (1. - self._decay) * v
        return self.z

    def next_eval(self, c: Callable[[], float]):
        v = c()
        return self.next(v)

    def end(self):
        return None



import pytest
from ml_instrumentation.Writer import Writer

@pytest.fixture
def writer():
    writer = Writer(
        db_path=':memory:',
        low_watermark=2,
        high_watermark=4,
    )
    yield writer
    writer.close()



def test_collector_setup1(basic_collector):
    assert basic_collector.keys() == set()
    assert basic_collector.experiment_ids() == set()

    # non-existent metrics and/or experiment_ids should give empty lists
    assert basic_collector.get('m1', 0) == []



import os
import time
import sqlite3
import logging
from collections import defaultdict
from typing import Any, NamedTuple
from concurrent.futures import ThreadPoolExecutor, Future

import filelock

import ml_instrumentation._utils.sqlite as sqlu

logger = logging.getLogger('ml-instrumentation')

class Point(NamedTuple):
    exp_id: int | str
    metric: str
    frame: int
    data: Any

class SqlPoint(NamedTuple):
    frame: int
    id: int | str
    measurement: Any

class Writer:
    def __init__(
        self,
        db_path: str,
        low_watermark: int = 64,
        high_watermark: int = 256,
    ):
        # ------------
        # -- config --
        # ------------
        self._db_path = db_path

        self._lw = low_watermark
        self._hw = high_watermark

        # ---------------
        # -- externals --
        # ---------------
        self._exec = ThreadPoolExecutor(max_workers=1)
        self._con = sqlite3.connect(self._db_path, check_same_thread=False)
        self._con.row_factory = row_factory

        self._write_future: Future | None = None

        # -----------
        # -- state --
        # -----------
        self._i = 0
        self._buffer: dict[str, dict[int, Point]] = defaultdict(dict)
        self._built: set[str] = set()

        self._avg_write_time = -1
        self._last_write_time = -1

        # -------------------------
        # -- load existing state --
        # -------------------------
        self._init_db()

    # ------------
    # -- IO API --
    # ------------
    def write(self, d: Point):
        self._buffer[d.metric][self._i] = d
        self._i += 1

        if self._i > self._hw:
            assert self._write_future is not None
            self._write_future.result()
            logger.warning(f'Buffer reached high watermark. (last write: {self._last_write_time}s, avg write: {self._avg_write_time}s)')
            self.sync()

        elif self._i > self._lw:
            self.sync()

    def sync(self):
        if self._write_future is not None and not self._write_future.done():
            return

        data = self._buffer
        self._buffer = defaultdict(dict)
        self._i = 0
        self._write_future = self._exec.submit(self._sync_async, data)

    def sync_now(self):
        if self._write_future is not None:
            self._write_future.result()

        self.sync()
        assert self._write_future is not None
        self._write_future.result()

    def read_metric(self, metric: str, exp_id: int | str | None = None) -> list[SqlPoint]:
        self.sync_now()

        cond = ''
        if exp_id is not None:
            cond = f'WHERE id={maybe_quote(exp_id)}'

        cur = self._con.cursor()
        try:
            cur = cur.execute(f'SELECT * FROM "{metric}" {cond}')
            res = cur.fetchall()
        except sqlite3.OperationalError:
            logger.warning(f'Specified metric/exp_id does not exist: <{metric}, {exp_id}>')
            res = []

        return res

    def dump(self) -> str:
        self._con.row_factory = None
        self.sync_now()
        return ''.join(self._con.iterdump())

    def close(self):
        self.sync_now()
        self._con.close()

    # -----------------
    # -- Utility API --
    # -----------------
    def metrics(self):
        buffered_keys = set(self._buffer.keys())
        return buffered_keys | self._built

    def merge(self, other: str):
        self.sync_now()

        with filelock.FileLock(f'{other}.lock'):
            other_con = sqlite3.connect(other)

            cur = self._con.cursor()
            other_cur = other_con.cursor()
            tables = sqlu.get_tables(cur)
            other_tables = sqlu.get_tables(other_cur)

            to_build = tables - other_tables
            for table in to_build:
                other_cur.execute(f'CREATE TABLE "{table}"(frame, id, measurement)')

            other_con.close()

            cur.execute(f'ATTACH DATABASE "{other}" AS other_db')
            for table in tables:
                cur.execute(f'INSERT INTO other_db.{table} SELECT * FROM {table}')

            self._con.commit()
            cur.execute('DETACH DATABASE other_db')
            cur.close()

    # -------------------
    # -- Backend logic --
    # -------------------
    def _sync_async(self, d: dict[str, dict[int, Point]]):
        sql_d = {}
        for m, sub in d.items():
            sql_d[m] = [
                SqlPoint(p.frame, p.exp_id, p.data) for p in sub.values()
            ]

        cur = self._con.cursor()
        start = time.perf_counter()

        for m in sql_d:
            self._setup_table(cur, m)
            self._write_many(cur, m, sql_d[m])

        self._con.commit()
        cur.close()
        elapsed = time.perf_counter() - start

        self._last_write_time = elapsed
        if self._avg_write_time < 0:
            self._avg_write_time = elapsed
        else:
            self._avg_write_time = 0.9 * self._avg_write_time + 0.1 * elapsed

    def _init_db(self):
        if not os.path.exists(self._db_path) and self._db_path != ':memory:':
            return

        cur = self._con.cursor()
        self._built |= sqlu.get_tables(cur)

    # ---------------------
    # -- utility methods --
    # ---------------------
    def _setup_table(self, cur: sqlite3.Cursor, name: str):
        if name in self._built:
            return

        cur.execute(f'CREATE TABLE "{name}"(frame INTEGER, id, measurement)')
        self._built.add(name)

    def _write_many(self, cur: sqlite3.Cursor, m: str, d: list[SqlPoint]):
        cur.executemany(f'INSERT INTO "{m}" (frame, id, measurement) VALUES (?,?,?)', d)


def row_factory(cur, d):
    return SqlPoint(*d)

def maybe_quote(x: int | str):
    if isinstance(x, str):
        return f'"{x}"'

    return x



from collections.abc import Callable, Sequence
import pytest
from collections import defaultdict
from typing import Any, NamedTuple
from ml_instrumentation.Collector import Collector
from ml_instrumentation.Sampler import Identity, Ignore
from ml_instrumentation.Writer import SqlPoint

@pytest.fixture
def basic_collector():
    c = Collector(
        tmp_file=':memory:',
        config = {
            'm1': Identity(),
            'm2': Identity(),
        },
        default=Ignore()
    )
    yield c
    c.close()


@pytest.fixture
def disk_collector(tmp_path):
    c = Collector(
        tmp_file=str(tmp_path / 'test.db'),
        config = {
            'm1': Identity(),
            'm2': Identity(),
        },
        default=Ignore()
    )
    yield c
    c.close()

# ----------------------
# -- Simulation Tools --
# ----------------------
class SimMetric(NamedTuple):
    metric: str
    pred: Callable[[int], bool]
    value: Callable[[int], Any]


def simulate_run(collector: Collector, start: int, end: int, metrics: Sequence[SimMetric]):
    expected: dict[str, list[SqlPoint]] = defaultdict(list)

    exp_id = collector.get_current_experiment_id()

    for i in range(start, end):
        collector.next_frame()

        for m in metrics:
            if m.pred(i):
                v = m.value(i)
                collector.collect(m.metric, v)
                expected[m.metric].append(SqlPoint(frame=i, id=exp_id, measurement=v))

    return expected



from collections.abc import Callable
from ml_instrumentation.Sampler import Sampler

class Pipe(Sampler):
    def __init__(self, *args: Sampler) -> None:
        self._subs = args

    def next(self, v: float) -> float | None:
        out: float | None = v
        for sub in self._subs:
            if out is None:
                return None
            out = sub.next(out)

        return out

    def next_eval(self, c: Callable[[], float]) -> float | None:
        subs = iter(self._subs)
        first = next(subs)
        out = first.next_eval(c)

        for sub in subs:
            if out is None:
                return None
            out = sub.next(out)

        return out

    def end(self):
        return None



from collections.abc import Callable
from typing import Any

import jax.experimental
from ml_instrumentation.Sampler import Sampler, Ignore, Identity, identity
from ml_instrumentation.Writer import Writer, Point

try:
    import jax
except ImportError:
    jax = None


class Collector:
    def __init__(
        self,
        tmp_file: str = ':memory:',
        config: dict[str, Sampler | Ignore] | None = None,
        default: Identity | Ignore = identity,
        experiment_id: int | str | None = None,
    ):
        self._c = config or {}

        self._ignore = set(k for k, sampler in self._c.items() if isinstance(sampler, Ignore))
        self._sampler: dict[str, Sampler] = {
            k: sampler for k, sampler in self._c.items() if not isinstance(sampler, Ignore)
        }

        self._tmp_file = tmp_file
        self._writer = Writer(
            db_path=self._tmp_file,
            low_watermark=1,
            high_watermark=2048,
        )

        self._exp_id = experiment_id
        self._frame: int = -1

        # create this once and cache it since it is stateless
        # avoid recreating on every step
        self._def = default

        # cache some useful metadata
        self._idxs = set[int | str]()
        self._keys = set[str]()

    # -------------
    # -- Context --
    # -------------
    def set_experiment_id(self, id: str | int):
        self._exp_id = id
        self._frame = -1

    def get_current_experiment_id(self) -> str | int:
        assert self._exp_id is not None
        return self._exp_id

    def get_frame(self):
        return self._frame

    def next_frame(self):
        self._frame += 1

    def set_frame(self, frame: int):
        assert frame >= self._frame
        self._frame = frame

    def reset(self):
        self.next_frame()
        for k in self._sampler:
            v = self._sampler[k].end()
            if v is None:
                continue

            self._write(k, v)

        self._frame = -1

    def close(self):
        self._writer.close()

    # -------------
    # -- Storing --
    # -------------
    def collect(self, name: str, value: Any):
        if name in self._ignore:
            return

        if self._frame == -1:
            self.next_frame()

        v = self._sampler.get(name, self._def).next(value)
        if v is None:
            return

        self._write(name, v)

    def evaluate(self, name: str, lmbda: Callable[[], Any]):
        if name in self._ignore:
            return

        if self._frame == -1:
            self.next_frame()

        v = self._sampler.get(name, self._def).next_eval(lmbda)
        if v is None:
            return

        self._write(name, v)

    def collect_jax(self, name: str, value: Any):
        if jax is None:
            raise Exception('jax is not installed')

        def _callback(metrics: dict[str, Any]):
            assert jax is not None
            for name, value in metrics.items():
                if isinstance(value, jax.Array):
                    value = value.item()

                self.collect(name, value)

        metrics = {name: value}
        jax.debug.callback(_callback, metrics, ordered=False)

    # ---------------
    # -- Accessing --
    # ---------------
    def get(self, metric: str, experiment_id: int | str):
        return self._writer.read_metric(metric, experiment_id)

    def keys(self):
        return self._keys

    def experiment_ids(self):
        return self._idxs

    def merge(self, loc: str):
        self._writer.merge(loc)

    # --------------
    # -- Internal --
    # --------------
    def _write(self, k: str, v: Any):
        self._keys.add(k)
        d = Point(
            exp_id=self.get_current_experiment_id(),
            metric=k,
            frame=self._frame,
            data=v,
        )
        self._writer.write(d)

    # -------------------
    # -- Serialization --
    # -------------------

    def __getstate__(self):
        self._writer.sync_now()

        data = None
        if self._tmp_file.startswith(':memory:'):
            data = self._writer.dump()

        ignore_keys = ['_writer']
        return {
            'data': data,
            'sub': {
                k: v for k, v in self.__dict__.items() if k not in ignore_keys
            },
        }

    def __setstate__(self, state):
        for k, v in state['sub'].items():
            self.__dict__[k] = v

        self._writer = Writer(
            db_path=self._tmp_file,
            low_watermark=1024,
            high_watermark=2048,
        )

        if state['data'] is not None:
            self._writer._con.executescript(state['data'])
            self._writer._init_db()



from ml_instrumentation.Writer import Writer, Point

def test_write1():
    writer = Writer(
        db_path=':memory:',
        low_watermark=2,
        high_watermark=4,
    )

    d = Point(
        metric='measurement-1',
        exp_id=0,
        frame=0,
        data=1.1,
    )
    writer.write(d)

    # should get back a list of datapoints for this measurement
    points = writer.read_metric('measurement-1')
    assert points == [
        (0, 0, 1.1),
    ]

    d = Point(
        metric='measurement-1',
        exp_id=0,
        frame=1,
        data=2.2
    )
    writer.write(d)

    points = writer.read_metric('measurement-1')
    assert points == [
        (0, 0, 1.1),
        (1, 0, 2.2),
    ]

    writer.close()

def test_write2():
    writer = Writer(
        db_path=':memory:',
        low_watermark=256,
        high_watermark=512,
    )

    for i in range(1_000):
        d = Point(
            metric='measurement-1',
            exp_id=0,
            frame=i,
            data=i * 1.1,
        )
        writer.write(d)

        if i % 3 == 0:
            d2 = d._replace(metric='measurement-2')
            writer.write(d2)

    points = writer.read_metric('measurement-1')
    assert len(points) == 1_000
    for i in range(1_000):
        assert points[i].frame == i
        assert points[i].measurement == i * 1.1

    points = writer.read_metric('measurement-2')
    assert len(points) == 334
